/*
 * Copyright (C) 2003 by Mark Roth
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the 
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* 
 * x125_cmyk.c
 *
 * Printer driver for the Lexmark X125 printer.
 * Portions copied from Alwin Stolk's Z42 driver and modified to work 
 * with the Lexmark X125 printer.
 *
 * X125 Author: Mark Roth
 * Original Z42 Author: Alwin Stolk
 * Contributor: Enrique Perez-Terron
 * Version: 0.2.3
 * Date: February 18, 2004
 * Copyright: (c) 2003-2004 by Mark Roth
 */

/***************************************************************************
gs -q -dBATCH -dPARANOIDSAFER -dQUIET -dNOPAUSE -sDEVICE=bitcmyk \
   -sPAPERSIZE=letter -r600x600 -sOutputFile=- - | \
   x125_cmyk size=letter hdpi=600 device=/dev/usb/lp0
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <fcntl.h>
#include <unistd.h>


#define HEADER_SZ	34
#define RAW_PACKET_SZ	28
#define PACKET_HDR_SZ	4

#define NOZZLES_BLACK	192
#define NOZZLES_COLOR   64

#define BLACK		1
#define YELLOW		2
#define MAGENTA		4
#define CYAN		8

#define BORDER_BOTTOM	0
#define BORDER_RIGHT	3

#define TRUE            1
#define FALSE           !TRUE

#define DEBUG_OUTPUT    FALSE // create a raw binary stream for debug?

#define BLACK_DELAY_LINES 32

enum printmodes {
	MODE_BLACK, MODE_CMY, MODE_CMYK
};


typedef unsigned char bool;

typedef struct {
	char *name;
	unsigned width;
	unsigned height;
} papersize;


char *modenames[] = { "k", "cmy", "cmyk", NULL };
unsigned char shift[] = 
        { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

/* PRN file Leadin (doesn't work) */
/*
#define LEADIN_LENGTH 48
unsigned char leadin[] = {
        0x1b, 0x2a, 0x81, 0x00, 0x1c, 0x56, 0x49, 0x00,
        0x01, 0x00, 0x58, 0x02, 0x01, 0x00, 0xc0, 0x12,
        0xc8, 0x19, 0x02, 0x00, 0x18, 0x00, 0x09, 0x00,
        0x06, 0x00, 0x08, 0x00, 0x1b, 0x2a, 0x07, 0x73,
        0x30, 0x1b, 0x2a, 0x6d, 0x00, 0x14, 0x01, 0xf4,
        0x02, 0x00, 0x01, 0xf0, 0x1b, 0x2a, 0x07, 0x63
};
*/
/* XXX - This can definitely be made shorter. */
#define LEADIN_LENGTH 0x1B3
unsigned char leadin[] = {
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 
        0x04, 0x01, 0x0f, 0x00, 0x00, 0x00, 0x00, 0xeb, 
        0x00, 0x00, 0x0f, 0x00, 0x01, 0x00, 0x01, 0x03, 
        0x03, 0x00, 0x10, 0x00, 0x10, 0xff, 0xff, 0x04, 
        0x01, 0x0f, 0x00, 0x00, 0x00, 0x00, 0xeb, 0x00, 
        0x00, 0x0f, 0x00, 0x01, 0x00, 0x01, 0x02, 0x02, 
        0x00, 0x10, 0x00, 0x10, 0xff, 0xff, 0x04, 0x02, 
        0x15, 0x00, 0x00, 0x00, 0x00, 0xe4, 0x00, 0x00, 
        0x0f, 0x00, 0x01, 0x00, 0x01, 0x04, 0x04, 0x00, 
        0x10, 0x00, 0x10, 0xff, 0xff, 0x02, 0x02, 0x06, 
        0x00, 0x04, 0x80, 0x04, 0x01, 0x06, 0x00, 0x00, 
        0x00, 0x00, 0xf4, 0x04, 0x04, 0x06, 0x00, 0x18, 
        0x80, 0x04, 0x01, 0x2a, 0x00, 0x00, 0x00, 0x00, 
        0xd0, 0x02, 0x02, 0x2a, 0x00, 0x00, 0x00, 0x1b, 
        0x70, 0x64, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x04, 0x01, 0x06, 0x00, 0x00, 
        0x00, 0x00, 0xf4, 0x02, 0x02, 0x06, 0x00, 0x01, 
        0x80, 0x04, 0x01, 0x14, 0x00, 0x00, 0x00, 0x00, 
        0xe6, 0x02, 0x02, 0x14, 0x00, 0x00, 0x00, 0x1b, 
        0x50, 0x61, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x06, 
        0x00, 0x00, 0x00, 0x00, 0xf4, 0x02, 0x02, 0x06, 
        0x00, 0x01, 0x80, 0x04, 0x01, 0x0e, 0x00, 0x00, 
        0x00, 0x00, 0xec, 0x02, 0x02, 0x0e, 0x00, 0x00, 
        0x00, 0x1b, 0x50, 0x4c, 0x00, 0x00, 0x00, 0x00, 
        0x00, 0x04, 0x01, 0x06, 0x00, 0x00, 0x00, 0x00, 
        0xf4, 0x02, 0x02, 0x06, 0x00, 0x01, 0x80, 0x04, 
        0x01, 0x0e, 0x00, 0x00, 0x00, 0x00, 0xec, 0x04, 
        0x04, 0x0e, 0x00, 0x00, 0x00, 0x1b, 0x50, 0x42, 
        0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x17, 
        0x00, 0x00, 0x00, 0x00, 0xe3, 0x04, 0x04, 0x17, 
        0x00, 0x01, 0x00, 0x1b, 0x70, 0x77, 0x09, 0x00, 
        0x00, 0x00, 0x00, 0xa5, 0x00, 0x06, 0x50, 0xe0, 
        0x73, 0x01, 0x01, 0x07, 0x04, 0x03, 0x54, 0x00, 
        0x00, 0x00, 0x00, 0xa4, 0x04, 0x04, 0x17, 0x00, 
        0x00, 0x00, 0x1b, 0x70, 0x77, 0x09, 0x00, 0x00, 
        0x00, 0x00, 0xa5, 0x00, 0x06, 0x40, 0x05, 0x00, 
        0x00, 0x00, 0x00, 0x04, 0x04, 0x2a, 0x00, 0x00, 
        0x00, 0x1b, 0x70, 0x77, 0x1c, 0x00, 0x00, 0x00, 
        0x00, 0x1b, 0x2a, 0x81, 0x00, 0x1c, 0x56, 0x49, 
        0x00, 0x01, 0x00, 0x58, 0x02, 0x01, 0x00, 0xc0, 
        0x12, 0xc8, 0x19, 0x02, 0x00, 0x18, 0x00, 0x09, 
        0x00, 0x05, 0x00, 0x08, 0x00, 0x04, 0x04, 0x13, 
        0x00, 0x00, 0x00, 0x1b, 0x70, 0x77, 0x05, 0x00, 
        0x00, 0x00, 0x00, 0x1b, 0x2a, 0x07, 0x73, 0x30,
        0x04, 0x02, 0x2B, 0x00, 0x00, 0x00, 0x00, 0xCE, 
        0x04, 0x04, 0x19, 0x00, 0x00, 0x00, 0x1B, 0x70, 
        0x77, 0x0b, 0x00, 0x00, 0x00, 0x00, 0x1B, 0x2A,
        0x6D, 0x00, 0x14, 0x01, 0xF4, 0x02, 0x00, 0x01,
        0xF0, 0x04, 0x04, 0x12, 0x00, 0x00, 0x00, 0x1B, 
        0x70, 0x77, 0x04, 0x00, 0x00, 0x00, 0x00, 0x1B, 
        0x2A, 0x07, 0x63
};

/* Leadmiddle (between pages) */
#define LEADMIDDLE_LENGTH 24
unsigned char leadmiddle[] = {
        0x1b, 0x2a, 0x07, 0x65, 0x1b, 0x2a, 0x07, 0x73,
        0x30, 0x1b, 0x2a, 0x6d, 0x00, 0x14, 0x01, 0xf4,
        0x02, 0x00, 0x01, 0xf0, 0x1b, 0x2a, 0x07, 0x63
};


/* Ideal Leadout for X125 (doesn't work) */
/*
#define LEADOUT_LENGTH 12
unsigned char leadout[] = {
        0x1b, 0x2a, 0x07, 0x65, 0x1b, 0x2a, 0x82, 0x00,
        0x00, 0x00, 0x00, 0xac
};
*/
/* Leadout for X125 */
#define LEADOUT_LENGTH 0xA2
unsigned char leadout[] = {
        0x04, 0x02, 0x28, 0x00, 0x00, 0x00, 0x00, 0xD1,
        0x04, 0x04, 0x12, 0x00, 0x00, 0x00, 0x1B, 0x70, 
        0x77, 0x04, 0x00, 0x00, 0x00, 0x00, 0x1B, 0x2A, 
        0x07, 0x65, 0x04, 0x04, 0x16, 0x00, 0x00, 0x00,
        0x1B, 0x70, 0x77, 0x08, 0x00, 0x00, 0x00, 0x00,
        0x1B, 0x2A, 0x82, 0x00, 0x00, 0x00, 0x00, 0xAC,
        0x04, 0x01, 0x19, 0x00, 0x00, 0x00, 0x00, 0xe1, 
        0x04, 0x04, 0x19, 0x00, 0x00, 0x00, 0x1b, 0x70, 
        0x77, 0x0b, 0x00, 0x00, 0x00, 0x00, 0xa5, 0x00, 
        0x08, 0x50, 0x05, 0x02, 0x00, 0x00, 0x00, 0x00, 
        0x01, 0x04, 0x01, 0x17, 0x00, 0x00, 0x00, 0x00, 
        0xe3, 0x04, 0x04, 0x17, 0x00, 0x00, 0x00, 0x1b, 
        0x70, 0x77, 0x09, 0x00, 0x00, 0x00, 0x00, 0xa5, 
        0x00, 0x06, 0x40, 0xe0, 0x73, 0x01, 0x01, 0x00, 
        0x04, 0x01, 0x0e, 0x00, 0x00, 0x00, 0x00, 0xec, 
        0x04, 0x04, 0x0e, 0x00, 0x00, 0x00, 0x1b, 0x50, 
        0x45, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x02, 
        0x14, 0x00, 0x00, 0x00, 0x00, 0xe5, 0x02, 0x02, 
        0x0e, 0x00, 0x00, 0x00, 0x1b, 0x50, 0x55, 0x00, 
        0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x06, 0x00, 
        0x01, 0x80
};

// units = pixels @ 72 dpi
// width  3"..8.5"
// height 5"..17"
papersize ghostscript[] = {
	{ "a4", 595, 842 },
	{ "a5", 421, 595 },
	{ "a6", 297, 421 },
	{ "legal", 612, 1008 },
	{ "letter", 612, 792 },
	{ "isob5", 501, 709 },
	{ "isob6", 354, 501 },
	{ "jisb5", 516, 729 },
	{ "jisb6", 363, 516 },
	{ "c5", 459, 649 },
	{ "c6", 323, 459 },
	{ "flsa", 612, 936 },
	{ "flse", 612, 936 },
	{ "halfletter", 396, 612 },
	{ NULL, 0, 0 }
};

int device=0;
FILE *in=NULL;
unsigned gs_width=595, gs_height=842; // default papersize is a4
unsigned hdpi=600; // default horizontal dpi
unsigned gap; // distance between the two rows of nozzles
bool bidirectional=FALSE;
short mode=MODE_BLACK;

unsigned char *chunk_buffer = 0;
long chunk_buffer_max = 0;
long chunk_buffer_size = 0;
int move_paper_amount = 0;

#define READ_BUFFER_SIZE 128
unsigned char *read_buffer = 0;

/***************************************************************************/

void convert_dpi(unsigned *val, unsigned from, unsigned to)
{
	float x;

	x = (float)*val * to / from;
	*val = (unsigned) x;
	x -= *val;
	if (x >= 0.5f) (*val)++;
}

/*
 * Writes data to the printer verbatim.  Also checks to see if there
 * are any characters the printer wants to send to us so we can clear
 * its buffer.  If we don't clear the buffer, large print jobs (2 or more
 * pages) will lock up the printer since the buffer is not flushed.
 */
void put_block_verbatim(char *p, long len)
{
        fd_set rfd, wfd;
        struct timeval tv;
        int count;
        
        if( read_buffer == 0 ) {
                read_buffer = malloc( READ_BUFFER_SIZE );
        }
        
        FD_ZERO( &rfd );
        FD_ZERO( &wfd );
        
        while( 1 ) {
                FD_SET( device, &rfd );
                FD_SET( device, &wfd );
                tv.tv_sec = 1;
                tv.tv_usec = 0;
                
                if( select( device + 1, &rfd, &wfd, NULL, &tv ) == -1 ) {
                        perror( "Error communicating with device" );
                        exit( 1 );
                }
                
                if( FD_ISSET( device, &rfd ) ) {
                        // clear printer input buffer
                        count = read( device, read_buffer, READ_BUFFER_SIZE );
                }
                
                if( FD_ISSET( device, &wfd ) ) {
                        if( (count = write( device, p, len )) <= 0 ) {
                                perror( "Error writing to device" );
                                exit( 1 );
                        }
                        break;
                }
        }
}

/*
 * Dynamically grows the chunk buffer array to handle a capacity of new_size.
 * If the array is large enough already, nothing happens.  If the buffer
 * does not yet exist, it is created.
 */
void grow_chunk_buffer( long new_size ) {
        unsigned char *new_chunk_buffer;
        
        if( chunk_buffer == 0 ) {
                // start at new_size
                chunk_buffer = malloc( new_size );
                chunk_buffer_max = new_size;
        }
        
        // Dynamically grow chunk buffer, if necessary:
        if( new_size > chunk_buffer_max ) {
                new_chunk_buffer = malloc( new_size );
                memcpy( new_chunk_buffer, chunk_buffer, chunk_buffer_size );
                free( chunk_buffer );
                chunk_buffer = new_chunk_buffer;
                chunk_buffer_max = new_size;
        }
}

/*
 * Prepares data to be sent to the output stream.  No data is actually
 * sent until commit_block() is called.  This is so that we can correctly
 * package the data block into 256-byte chunks for the X125.
 * Always commit before starting a new command.
 */
void put_block(char *p, long len)
{
        grow_chunk_buffer( chunk_buffer_size + len );
        memcpy( chunk_buffer + chunk_buffer_size, p, len );
        chunk_buffer_size += len;
}

/*
 * Commits the data sent to put_block to the stream.  Packages into 
 * 256-byte chunks for the X125.  Assumes we are committing a single
 * command.
 */
void commit_block() {
        unsigned char block_header[] = 
        //             count, len0, len1, len2, len3,  pad, cksum
                { 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, };
        unsigned long block_len;
        unsigned char command_header[] =
        //                    cmdlen
                { 0x04, 0x04, 0x00, 0x00, 0x00, 0x00, 
        //                          size
                  0x1B, 0x70, 0x77, 0x00, 0x00, 0x00, 0x00, 0x00};
        unsigned short command_len;
        short count, blocks, remainder, i, j;
        long remaining;
        unsigned char *buf = chunk_buffer;
                
        remaining = chunk_buffer_size;
        while( remaining > 0 ) {
                // Determine how many blocks to encode this time
                // (at most 24 at a time to immitate windows driver)
                // Each command can hold at most 256 bytes
                blocks = ((remaining-1) / 256) + 1;
                if( blocks > 24 ) blocks = 24;
                block_header[1] = blocks;
                
                // Determine block length
                // Is last command < 256 bytes?
                if( (blocks * 256) > remaining ) {
                        block_len = (14+256) * (blocks-1) + 14 + 
                                (remaining & 0xFF);
                }
                else {
                        block_len = (14+256) * blocks;
                }
                block_header[2] = block_len & 0xFF;
                block_header[3] = (unsigned char)(block_len >> 8);
                block_header[4] = (unsigned char)(block_len >> 16);
                block_header[5] = (unsigned char)(block_len >> 24);
                block_header[7] = 0xFF - block_header[0] - block_header[1] -
                    block_header[2] - block_header[3] - block_header[4] -
                    block_header[5] - block_header[6];
                put_block_verbatim( block_header, 8 );

                while( blocks > 0 ) {
                        if( remaining >= 256 ) {
                                command_len = 14 + 256;
                        }
                        else {
                                command_len = 14 + remaining;
                        }
                        command_header[2] = command_len & 0xFF;
                        command_header[3] = command_len >> 8;
                        command_header[9] = (command_len-14) & 0xFF;
                        
                        command_len -= 14;
                        put_block_verbatim( command_header, 14 );
                        put_block_verbatim( buf, command_len );
                        
                        buf += command_len;
                        remaining -= command_len;
                        blocks--;
                }
        }

        chunk_buffer_size = 0;
}


void crop_buffer( unsigned char *p, unsigned *packets, unsigned *left )
{
        unsigned x, y;
        unsigned right = 0;
        unsigned char leftset = 0;
        
        // find left
        for( x = 0; x < *packets; x++ ) {
                for( y = 0; y < 24; y++ ) {
                        if( p[x*24+y] ) {
                                right = x;
                                if( !leftset ) {
                                     *left = x;
                                     leftset = 1;
                                }
                        }
                }
        }
        
        if( leftset ) {
                *packets = right - *left;
        }
        else {
                *packets = 0;
                *left = 0;
        }
}

long make_packets_forward( unsigned char *dest, unsigned char *src,
        unsigned packets, unsigned left )
{
        unsigned char *dir;
        unsigned x, y;
        long length;
        unsigned short t, i, bt;
        unsigned short test;
        unsigned short directory;
        unsigned short bits;
        
        length = 0;
        test = 0x80;
        
        src += left * 24;
        for( x = left; x < (left+packets); x++ ) {
                // build directory, start with lowest-order bit
                directory = 0x2FFF;
                dir = dest;
                dest += 2;
                length += 2;
                for( t = 0x1; t < 0x1000; t <<= 1 ) {
                        // get next 16 bits
                        bits = 0;
                        bt = 0x8000;
                        for( i = 0; i < 16; i++ ) {
                                if( (*src) & test ) {
                                        bits |= bt;
                                }
                                bt >>= 1;
                                test >>= 1;
                                if( test == 0 ) {
                                        test = 0x80;
                                        src++;
                                }
                        }
                        if( bits ) {
                                // clear bit in directory
                                directory = directory & (0xFFFF - t);
                                *(dest++) = bits >> 8;
                                *(dest++) = bits & 0xFF;
                                length += 2;
                        }
                }
                *(dir) = (directory >> 8);
                *(dir+1) = directory & 0xFF;
        }
        
        return length;
}

long make_packets_reverse( unsigned char *dest, unsigned char *src,
        unsigned packets, unsigned left )
{
        unsigned char *dir;
        int x, y;
        long length;
        unsigned short t, i, bt;
        unsigned short test;
        unsigned short directory;
        unsigned short bits;
        
        length = 0;
        test = 0x80;
        
        src += (left+packets-1) * 24;
        for( x = (left+packets-1); x >= left; x-- ) {
                // build directory, start with lowest-order bit
                directory = 0x2FFF;
                dir = dest;
                dest += 2;
                length += 2;
                for( t = 0x1; t < 0x1000; t <<= 1 ) {
                        // get next 16 bits
                        bits = 0;
                        bt = 0x8000;
                        for( i = 0; i < 16; i++ ) {
                                if( (*src) & test ) {
                                        bits |= bt;
                                }
                                bt >>= 1;
                                test >>= 1;
                                if( test == 0 ) {
                                        test = 0x80;
                                        src++;
                                }
                        }
                        if( bits ) {
                                // clear bit in directory
                                directory = directory & (0xFFFF - t);
                                *(dest++) = bits >> 8;
                                *(dest++) = bits & 0xFF;
                                length += 2;
                        }
                }
                *(dir) = (directory >> 8);
                *(dir+1) = directory & 0xFF;
                src -= 48;
        }
        
        return length;
}

long make_packets( unsigned char *dest, unsigned char *src, 
        unsigned packets, unsigned left, unsigned reverse ) 
{
        if( reverse ) {
                return make_packets_reverse( dest, src, packets, left );
        }
        else {
                return make_packets_forward( dest, src, packets, left );
        }
}

void commit_move_paper() {
	unsigned char move[] = { 0x1B,0x2A,0x03,0x01,0x20 };

        if( move_paper_amount > 0 ) {
                move[3] = (char)(move_paper_amount>>8);
                move[4] = (char) move_paper_amount;
                put_block(move, 5);
                commit_block();
                move_paper_amount = 0;
        }
}
	
void move_paper(int dpi1200)
{
        move_paper_amount += dpi1200;
}


void make_header(unsigned char *header, unsigned long blk_length, 
        unsigned packets, unsigned hstart, unsigned yoff, bool color, 
        unsigned reverse )
{
	unsigned hend;

	header[0x00] = 0x1B; // magic 0
	header[0x01] = 0x2A; // magic 1
	header[0x02] = 0x24; // print command

	header[0x03] = (char)(blk_length >> 24);
	header[0x04] = (char)(blk_length >> 16);
	header[0x05] = (char)(blk_length >> 8);
	header[0x06] = (char)(blk_length);

	switch (hdpi) {
		case 300: header[7]=1; break;
		case 600: header[7]=3; break;
		case 1200: header[7]=4; break;
		case 2400: header[7]=5; break;
	}

        // direction: 1=>left2right 2=>right2left
        if( reverse ) {
                header[0x08] = 2;
        }
        else {
                header[0x08] = 1; 
        }

	if (color) {
		header[0x09] = 0x00;
	} else {
		header[0x09] = 0x01;
	}

	header[0x0A] = 0x18; // nozzles: 0x18=>192 0x1A=>208
	header[0x0B] = 0;
	header[0x0C] = 1;    // ?

	header[0x0D] = (char)(packets >> 8);// (packets-1)*4 = hend-hstart
	header[0x0E] = (char)(packets);

	header[0x0F] = (char)(hstart >> 8); // horizontal start BE
	header[0x10] = (char)(hstart);

	hend = hstart + (packets-1)*(2400/hdpi);
	header[0x11] = (char)(hend >> 8); // horizontal end BE
	header[0x12] = (char)(hend);

        header[0x13] = (char)(yoff >> 8); // y offset BE (1200 dpi)
        header[0x14] = (char)(yoff);

	header[0x15] = 0;
	header[0x16] = 0;
        
	header[0x17] = 0;
	header[0x18] = 0;
	header[0x19] = 0;
	header[0x1A] = 0;
	header[0x1B] = 0;
	header[0x1C] = 0;
	header[0x1D] = 0;
	header[0x1E] = 0;
	header[0x1F] = 0;
	header[0x20] = 0;
	header[0x21] = 0;
}

/*
 * Make a color stripe and clear out the bits in the component stripes.
 */
void make_color_stripe( unsigned char *bufcmy, unsigned char *bufc, 
        unsigned char *bufm, unsigned char *bufy, long y, 
        unsigned xmax ) 
{
        unsigned short y0 = y%248, i, ei, ey;
        long x;
        
        for( x = 0; x < xmax; x++ ) {
                for( i = 0; i < NOZZLES_COLOR; i++ ) {
                        ey = (y0+i)%248;
                        // C = pins 0-63
                        ei = i;
                        if( bufc[31*x + ey/8] & shift[ey&7] ) {
                                bufcmy[24*x + ei/8] |= shift[ei&7];
                                bufc[31*x + ey/8] &= (255-shift[ey&7]);
                        }
                        
                        // M = pins 64-127
                        ei = 64 + i;
                        if( bufm[31*x + ey/8] & shift[ey&7] ) {
                                bufcmy[24*x + ei/8] |= shift[ei&7];
                                bufm[31*x + ey/8] &= (255-shift[ey&7]);
                        }
                        
                        // Y = pins 128-191
                        ei = 128 + i;
                        if( bufy[31*x + ey/8] & shift[ey&7] ) {
                                bufcmy[24*x + ei/8] |= shift[ei&7];
                                bufy[31*x + ey/8] &= (255-shift[ey&7]);
                        }
                }
        }
}

void print(void)
{
        unsigned width, height;
        long i, x, ex, y, ey, buffer_size, color_buffer_size, length;
        unsigned packets, left;
        int bidi_left, bidi_right, bidi_last;
        unsigned reverse;
        unsigned char *bufk, *black_delay;
        unsigned char *bufc, *bufm, *bufy, *bufcmy;
        unsigned char *packet_buffer;
        unsigned char *lbuf;
        unsigned hgap, yoff, black_delay_y = 0;
	unsigned char pixel, header[HEADER_SZ];
        FILE *ftmp;
        unsigned char leadin_sent = 0;
        
	// set up resolution specific parameters
	switch (hdpi) {
		case 300: hgap=8; break;
		case 600: hgap=16; break;
		case 1200: hgap=32; break;
		case 2400: hgap=64; break;
	}
        
        if( DEBUG_OUTPUT ) {
                ftmp = fopen( "/tmp/output2.bin", "w" );
        }
        
        // set up width and height
	width = gs_width - BORDER_RIGHT;
	height = gs_height - BORDER_BOTTOM;
	convert_dpi(&width, 72, hdpi);
	convert_dpi(&height, 72, 600);
	convert_dpi(&gs_width, 72, hdpi);
	convert_dpi(&gs_height, 72, 600);
	if (width&1) width++;	// ghostscript has even lines. this is good.
	if (gs_width&1) gs_width++;

        // allocate memory for stripes
        // (192 nozzles = 192 rows / 8 bits per byte = 24 bytes per column)
        buffer_size = 24 * (width + hgap + 1);
        bufk = malloc( buffer_size );
        memset( bufk, 0, buffer_size );
        
        // (64 cyan nozzles + 28 spaces + 64 magenta nozzles + 28 spaces + 
        //  64 yellow nozzles = 248 scan lines)
        // (248 scan lines / 8 bits per byte = 31 bytes per column)
        if( mode != MODE_BLACK ) {
                color_buffer_size = 31 * (width + hgap + 1);
                bufc = malloc( color_buffer_size );
                memset( bufc, 0, color_buffer_size );
                bufm = malloc( color_buffer_size );
                memset( bufm, 0, color_buffer_size );
                bufy = malloc( color_buffer_size );
                memset( bufy, 0, color_buffer_size );
                
                bufcmy = malloc( buffer_size );
                memset( bufcmy, 0, buffer_size );
        }
        
        if( mode == MODE_CMYK ) {
                // If we're printing in CMYK, black needs to be offset by 64
                // Here, we set up a 64 scanline delayed-buffer:
                black_delay = malloc( gs_width / 2 * BLACK_DELAY_LINES );
        }
        
        // Buffer to read CMYK bytes
        lbuf = malloc( gs_width / 2 );
        
        // packet buffer
        // add another 2 bytes for directory during compression (worst case)
        packet_buffer = malloc( 26 * (width + hgap + 1) );
        
	// print
	while( 1 ) {
                y=0;

                if( mode == MODE_CMYK ) {
                        memset( black_delay, 0, 
                                gs_width / 2 * BLACK_DELAY_LINES );
                }
                
                fgetc(in);
                if( feof(in) ) break;
                fseek(in, -1, SEEK_CUR);

                if( !leadin_sent ) {
                        put_block_verbatim( leadin, LEADIN_LENGTH );
                        //put_block( leadin, LEADIN_LENGTH );
                        //commit_block();
                        leadin_sent = 1;
                }
                else {
                        put_block( leadmiddle, LEADMIDDLE_LENGTH );
                        commit_block();
                }
                
                // seems to be the right amount to get top margin right.
                if( (mode != MODE_BLACK) && !(y%NOZZLES_COLOR) ) {
        		move_paper( 0x40 );
                }
                else {
        		move_paper( 0x180 );
                }

                bidi_last = 0;
		while( y < height ) {
                        // gs encodes cmykcmyk in one byte (2 pixels per byte)
			fread(lbuf, 1, gs_width/2, in);
                        
			// pixel line
			for (x=0; x<width; x++) {
                                // Adjust effective x by hgap if even row.
                                // This compensates for the gap between 
                                // odd/even pins
                                ex = (y&1) ? x : (x+hgap);
                                
				pixel = lbuf[x>>1];
				if (x&1) {
					pixel &= 15;
				} else {
					pixel >>= 4;
				}

				switch (mode) {
					case MODE_BLACK:
                                                if( pixel ) {
                                                        pixel = BLACK;
                                                }
						break;

					case MODE_CMY:
						// fake black
						if (pixel&BLACK) {
                                                        pixel = CYAN|MAGENTA|
                                                                YELLOW;
                                                }
						break;
                                        case MODE_CMYK:
                                                // treat CMY as K
                                                if( (pixel&(CYAN|MAGENTA|
                                                        YELLOW)) == 
                                                        (CYAN|MAGENTA|
                                                        YELLOW) ) 
                                                {
                                                        pixel = BLACK;
                                                }
                                                break;
				}
                                
                                if( mode != MODE_BLACK ) {
                                        // Colors
                                        // store in stripe
                                        if( pixel & CYAN ) {
                                                ey = y%248;
                                                bufc[31*ex + (ey/8)] |=
                                                        shift[ey&7];
                                        }
                                        if( pixel & MAGENTA ) {
                                                ey = (y+156)%248;
                                                bufm[31*ex + (ey/8)] |=
                                                        shift[ey&7];
                                        }
                                        if( pixel & YELLOW ) {
                                                ey = (y+64)%248;
                                                bufy[31*ex + (ey/8)] |=
                                                        shift[ey&7];
                                        }
                                        
                                        if( mode == MODE_CMYK ) {
                                                // In CMYK mode, we need to 
                                                // figure out black based on 
                                                // 64-scanline delay.
                                                pixel = (black_delay + 
                                                        black_delay_y * 
                                                        gs_width / 2)
                                                        [x>>1];
                                                if (x&1) {
                                                        pixel &= 15;
                                                } else {
                                                        pixel >>= 4;
                                                }
                                        
                                                // treat CMY as K
                                                if( (pixel&(CYAN|MAGENTA|
                                                        YELLOW)) == 
                                                        (CYAN|MAGENTA|
                                                        YELLOW) ) 
                                                {
                                                        pixel = BLACK;
                                                }
                                                
                                                if( pixel & BLACK ) {
                                                        bufk[24*ex + ((y%
                                                                NOZZLES_BLACK)
                                                                /8)] |=
                                                                shift[y&7];
                                                }
                                        }
                                }
                                else {
                                        if (pixel & BLACK) {
                                                // store in stripe
                                                bufk[24*ex + ((y%
                                                        NOZZLES_BLACK)/8)] |= 
                                                        shift[y&7];
                                        }
                                }
			}

                        if( mode == MODE_CMYK ) {
                                // If CMYK, copy current line into delay buffer
                                memcpy( black_delay + black_delay_y * 
                                        gs_width / 2, lbuf, 
                                        gs_width / 2 );
                                black_delay_y = (black_delay_y + 1) % 
                                        BLACK_DELAY_LINES;
                        }
                        
			y++;
			if (y==height) {
                                // Print last color stripe
                                while( y%NOZZLES_COLOR ) y++;
			}
                        
                        if( (mode != MODE_BLACK) && !(y%NOZZLES_COLOR) ) {
                                // Render color stripe
                                left = 0;
                                packets = gap + width;
                                make_color_stripe( bufcmy, 
                                        bufc, bufm, bufy, y, width + hgap + 1 );
                                crop_buffer( bufcmy, &packets, &left );
                                if( !(y%NOZZLES_BLACK) ) {
                                        yoff = 0;
                                }
                                else {
                                        yoff = NOZZLES_COLOR * 2;
                                }
                                if( packets == 0 ) {
                                        move_paper( yoff );
                                }
                                else {
                                        reverse = abs(left - bidi_last) > 
                                                abs((left+packets) - bidi_last);
                                        bidi_last = reverse ? left : 
                                                (left+packets);
                                        length = HEADER_SZ + make_packets( 
                                                packet_buffer, bufcmy, 
                                                packets, left, reverse );
                                        make_header( header, length, packets, 
                                                0x10+(left*(2400/hdpi)), yoff,
                                                TRUE, reverse );
                                        commit_move_paper();
                                        put_block( header, HEADER_SZ );
                                        put_block( packet_buffer, 
                                                length-HEADER_SZ );
                                        commit_block();
                                        
                                        if( DEBUG_OUTPUT ) {
                                                fwrite( header, 1, HEADER_SZ, 
                                                        ftmp );
                                                fwrite( packet_buffer, 1, 
                                                        length-HEADER_SZ, 
                                                        ftmp );
                                        }
                                }
                                memset( bufcmy, 0, buffer_size );
                        }

			if (y>=height) {
                                // Print last black stripe
                                while( y%NOZZLES_BLACK ) y++;
			}
                        
			if ( !(y%NOZZLES_BLACK) ) {
                                left = 0;
                                packets = gap + width;
                                crop_buffer( bufk, &packets, &left );
                                if( mode != MODE_BLACK ) {
                                        yoff = NOZZLES_COLOR * 2;
                                }
                                else {
                                        yoff = NOZZLES_BLACK * 2;
                                }
                                if( packets == 0 ) {
                                        move_paper( yoff );
                                }
                                else {
                                        reverse = abs(left - bidi_last) > 
                                                abs((left+packets) - bidi_last);
                                        bidi_last = reverse ? left : 
                                                (left+packets);
                                        length = HEADER_SZ + make_packets( 
                                                packet_buffer, bufk, 
                                                packets, left, reverse );
                                        make_header( header, length, packets, 
                                                0x10+(left*(2400/hdpi)), yoff,
                                                FALSE, reverse );
                                        commit_move_paper();
                                        put_block( header, HEADER_SZ );
                                        put_block( packet_buffer, 
                                                length-HEADER_SZ );
                                        commit_block();
                                        
                                        if( DEBUG_OUTPUT ) {
                                                fwrite( header, 1, HEADER_SZ, 
                                                        ftmp );
                                                fwrite( packet_buffer, 1, 
                                                        length-HEADER_SZ, 
                                                        ftmp );
                                        }
                                }
                                memset( bufk, 0, buffer_size );
                        }
		}

                if (height<gs_height) {
                        for (i=0; i<(gs_height-height); i++) 
                                fread(lbuf, 1, gs_width/2, in); 
                }

                move_paper_amount = 0;
	}
d:
        
        put_block_verbatim(leadout, LEADOUT_LENGTH);
        //put_block(leadout, LEADOUT_LENGTH);
        //commit_block();
        
        if( DEBUG_OUTPUT ) {
                fclose( ftmp );
        }
        
        if( mode != MODE_BLACK ) {
                free( bufcmy );
                free( bufy );
                free( bufm );
                free( bufc );
        }
        if( mode == MODE_CMYK ) {
                free( black_delay );
        }
        free( bufk );
        free( packet_buffer );
        free( lbuf );
}


char *find_param(char *p, int *nr)
{
	char *q, *parameters[] = {"size", "hdpi", "mode", "bidi", "device", 
                NULL};
	int i;

	for (i=0; parameters[i]; i++) {
		if ( !strncmp(p, parameters[i], strlen(parameters[i])) ) {
			q = p + strlen(parameters[i]);
			if ( *q == '=') {
				*nr=i;
				return q+1;
			}
		}
	}

	*nr=-1;
	return NULL;
}


int main(int argc, char *argv[])
{
	int arg, i, parm;
	char *p, *devicename = 0;

	// process command line parameters
	for (arg=1; arg<argc; arg++) {
		p = find_param(argv[arg], &parm);
		switch(parm) {
			case 0:
				for (i=0; ghostscript[i].name; i++) {
					if ( !strcmp(ghostscript[i].name, p) ) {
						gs_width = ghostscript[i].width;
						gs_height = 
                                                        ghostscript[i].height;
						fprintf(stderr, 
                                                     "page size set to '%s'.\n",
                                                     p);
						break;
					}
				}
				break;
			
			case 1:
				i=atoi(p);
				if (i==300 || i==600 || i==1200) {
					hdpi=i;
					fprintf(stderr, "horizontal resolution"
                                                "set to %d dpi.\n", i);
				}
				break;

			case 2:
				for (i=0; modenames[i]; i++) {
					if ( !strcmp(modenames[i], p) ) {
						mode=i;
						fprintf(stderr, 
                                                        "mode set to '%s'.\n", 
                                                        modenames[i]);
					}
				}
				break;

			case 3:
				bidirectional=atoi(p);
				fprintf(stderr, "bidirectional mode %s.\n",
                                        bidirectional ? "on" : "off");
				break;
                        case 4:
                                devicename=p;
                                fprintf(stderr, "device set to %s.\n", 
                                        devicename);
                                break;
		}
	}

        if( !devicename ) {
                fprintf( stderr, "Error: device parameter not specified.\n" );
                exit( 1 );
        }
        
	// start printing
	in = stdin;
	device = open( devicename, O_RDWR );
        if( device <= 0 ) {
                perror( "Could not open device" );
        }
        
	if (in && device) print();

	if (in) fclose(in);
	if (device) close(device);
        
	return 0;
}

